import argparse
import logging
import re
from urllib.parse import urlparse

import boto3

log_format = "[%(levelname)s] %(message)s"
logging.basicConfig(level=logging.INFO, format=log_format)


def build_cache_full_uri(s3_uri):
    parsed = urlparse(s3_uri)
    bucket = parsed.netloc
    prefix = parsed.path.lstrip("/")

    # Initialize S3 client
    s3_client = get_aws_client("s3")

    # Use Delimiter to list "subfolders"
    response = s3_client.list_objects_v2(Bucket=bucket, Prefix=prefix, Delimiter="/")

    if "CommonPrefixes" not in response or len(response["CommonPrefixes"]) == 0:
        raise ValueError(f"No subfolders found under the prefix: {s3_uri}")
    if len(response["CommonPrefixes"]) > 1:
        raise ValueError(f"Multiple subfolders found under the prefix: {s3_uri}")

    # Retrieve the single subfolder
    subfolder = response["CommonPrefixes"][0]["Prefix"]
    # Construct and return the full S3 URI
    return f"s3://{bucket}/{subfolder}"


def get_run_cache_path(run_id, task_id=None):
    omics_client = get_aws_client("omics")

    omics_run = omics_client.get_run(id=run_id)
    logging.debug(f"omics run: {omics_run}")
    cache_id = omics_run["cacheId"]

    omics_cache = omics_client.get_run_cache(id=cache_id)
    logging.debug(f"omics run cache: {omics_cache}")
    cache_base_uri = f"{omics_cache['cacheS3Uri']}{run_id}"
    if not task_id:
        logging.info(f"Run cache uri: {cache_base_uri}")
        cache_uri = cache_base_uri
    else:
        task_cache_uri = f"{cache_base_uri}/{task_id}/"
        full_cache_uri = build_cache_full_uri(task_cache_uri)
        logging.info(f"Run task cache uri: {full_cache_uri}")
        cache_uri = full_cache_uri
    return cache_uri


def get_aws_client(service):
    return boto3.client(service)


def list_objects_with_index(s3_client, bucket_name, prefix):
    """
    List all objects under the given S3 prefix that contain '_index' in their keys.
    """
    paginator = s3_client.get_paginator("list_objects_v2")
    for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix):
        for obj in page.get("Contents", []):
            key = obj["Key"]
            if "_index" in key:
                yield key


def copy_object(s3_client, bucket_name, source_key, dest_key):
    """
    Copy an S3 object from source_key to dest_key within the same bucket.
    """
    copy_source = {"Bucket": bucket_name, "Key": source_key}
    s3_client.copy(copy_source, bucket_name, dest_key)
    logging.info(f"Copied {source_key} to s3://{bucket_name}/{dest_key}")


def process_s3_uri(s3_uri):
    """
    Parse the S3 URI and return the bucket name and prefix.
    """
    parsed = urlparse(s3_uri)
    if parsed.scheme != "s3":
        raise ValueError(f"Invalid S3 URI: {s3_uri}")
    return parsed.netloc, parsed.path.lstrip("/")


def copy_omics_cached_indexes(cache_s3_uri):
    # Initialize S3 client
    s3_client = get_aws_client("s3")

    # Parse bucket and prefix
    bucket_name, prefix = process_s3_uri(cache_s3_uri)

    # List and process objects
    for source_key in list_objects_with_index(s3_client, bucket_name, prefix):
        # Create destination key by removing '_index' substring
        dest_key = re.sub(r"_index", "", source_key)
        copy_object(s3_client, bucket_name, source_key, dest_key)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("run_id", help="AWS HealthOmics run id")
    parser.add_argument(
        "--task-id", type=str, help="HealthOmics run task-id. Leave empty to get the base run cache uri."
    )
    parser.add_argument("--copy-indexes", dest="copy_indexes", action="store_true", help="copy indexes cached files")
    parser.set_defaults(copy_indexes=False)
    args = parser.parse_args()

    cache_s3_uri = get_run_cache_path(args.run_id, args.task_id)
    if args.copy_indexes:
        copy_omics_cached_indexes(cache_s3_uri)


if __name__ == "__main__":
    main()



import json
import logging
from argparse import ArgumentParser

import boto3

logging.basicConfig(level=logging.INFO)

LAMBDA_FUNCTION_NAME = "omicsEndRunHandler"
SUCCESS_STATUS_CODE = 200


def report_run_completion(run_ids: list, aws_region):
    account_id = boto3.client("sts").get_caller_identity().get("Account")
    lambda_client = boto3.client("lambda", region_name=aws_region)
    for run_id in run_ids:
        lambda_input = {
            "detail-type": "Run Status Change (Manual)",
            "source": "ugbio_omics.report_run_completion",
            "account": account_id,
            "region": aws_region,
            "runAnalyzerResult": {"Payload": {"run_id": run_id}},
        }
        response = lambda_client.invoke(
            FunctionName=LAMBDA_FUNCTION_NAME, InvocationType="RequestResponse", Payload=json.dumps(lambda_input)
        )
        logging.info(f"{LAMBDA_FUNCTION_NAME} was triggered manually with event: {lambda_input}")

        if response["StatusCode"] != SUCCESS_STATUS_CODE:
            raise Exception("Error. StatusCode: " + response["StatusCode"])


def main(raw_args=None):
    parser = ArgumentParser(description="Reports completion of Omics Run by invoking omicsEndRunHandler lambda")
    parser.add_argument("run_ids", help="Omics run ids (space separated)", nargs="+")
    parser.add_argument("-r", "--aws_region", type=str, help="aws region", default="us-east-1")
    args = parser.parse_args(raw_args)
    report_run_completion(args.run_ids, args.aws_region)


if __name__ == "__main__":
    main()



from datetime import UTC, datetime

import boto3


def get_run_info(run_id, client=None, *, get_tasks=True):
    if not client:
        client = boto3.client("omics")

    run = client.get_run(id=run_id)
    run.update({"duration": (run["stopTime"] if "stopTime" in run else datetime.now(UTC)) - run["startTime"]})

    if get_tasks:
        response = client.list_run_tasks(id=run_id)
        tasks = response["items"]
        while response.get("nextToken"):
            response = client.list_run_tasks(id=run_id, startingToken=response.get("nextToken"))
            tasks += response["items"]

        def calc_task_duration(task):
            stop_time = task["stopTime"] if "stopTime" in task else datetime.now(UTC)
            start_time = (
                task["startTime"] if "startTime" in task else stop_time
            )  # when task is CANCELLED sometime there's no startTime
            return stop_time - start_time

        tasks = [{**task, "duration": calc_task_duration(task)} for task in tasks]
        run["tasks"] = tasks

    del run["ResponseMetadata"]

    return run



import os
from argparse import ArgumentParser
from datetime import timedelta, tzinfo
from enum import Enum

import boto3
import dateutil.parser
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from ugbio_omics.get_omics_log import OMICS_LOG_GROUP
from ugbio_omics.get_run_cost import PLOTS_DIR, RUN_ID_PLACEHOLDER
from ugbio_omics.get_run_info import get_run_info


class EST(tzinfo):
    def utcoffset(self, dt):
        return timedelta(hours=-5)

    def dst(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "EST"


class EDT(tzinfo):
    def utcoffset(self, dt):
        return timedelta(hours=-4)

    def dst(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "EDT"


tzinfos = {"EDT": EDT(), "EST": EST()}


class MonitorColumns(Enum):
    TIME = "time"
    CPU = "CPU"
    MEMORY = "Memory"
    IO_RKB = "IO_rKb/s"
    IO_WKB = "IO_wKb/s"
    IOWAIT = "IOWait"


class MonitorLog:
    def __init__(self):
        self.df_columns = [column.value for column in MonitorColumns]
        self.df = pd.DataFrame(columns=self.df_columns)
        self.start_time = None
        self.run_time = timedelta(0)
        self.total_cpu = None
        self.total_memory = None
        self.task_name = None
        self.task_id = None
        # create a dict where key is the column name and value is the color
        self.colors = dict(zip(self.df_columns, ["black", "blue", "green", "red", "purple", "orange"], strict=False))

    def process_line(self, line):  # noqa: C901, PLR0912
        if "MONITORING" not in line:
            return

        # Get general information
        if "General Information" in line:
            # line=MONITORING, [Tue Feb 13 15:33:19 IST 2024], General Information, CPU: 20, Memory(GiB): 7
            split_line = line.split("MONITORING")[1].split(",")
            # Convert the date string: Tue Feb 13 15:33:19 IST 2024
            date_str = split_line[1].split("[")[1].split("]")[0]
            time = dateutil.parser.parse(date_str, tzinfos=tzinfos)
            if self.start_time is None:
                self.start_time = time
            try:
                self.total_cpu = int(split_line[3].split(":")[1] or 0)
            except ValueError:
                self.total_cpu = 0
            try:
                self.total_memory = float(split_line[4].split(":")[1].split("GiB")[0] or 0)
            except ValueError:
                self.total_memory = 0.0

        # Get monitoring information
        else:
            # line=MONITORING, [Sun Mar 24 21:35:30 UTC 2024], %CPU: 52.10, %Memory: 21.00, IO_rKb/s: 41.00, IO_wKb/s: 367.00, %IOWait: 0.00  # noqa: E501
            split_line = line.split("MONITORING")[1].split(",")
            # Convert the date string: Tue Feb 13 15:33:28 IST 2024
            date_str = split_line[1].split("[")[1].split("]")[0]
            time = dateutil.parser.parse(date_str, tzinfos=tzinfos)
            try:
                cpu = float(split_line[2].split(":")[1] or 0)
            except (ValueError, IndexError):
                cpu = 0.0
            try:
                memory = float(split_line[3].split(":")[1] or 0)
            except (ValueError, IndexError):
                memory = 0.0
            try:
                io_rkb = float(split_line[4].split(":")[1] or 0)
            except (ValueError, IndexError):
                io_rkb = 0.0
            try:
                io_wkb = float(split_line[5].split(":")[1] or 0)
            except (ValueError, IndexError):
                io_wkb = 0.0
            try:
                iowait = float(split_line[6].split(":")[1] or 0)
            except (ValueError, IndexError):
                iowait = 0.0

            if self.start_time is None:
                self.start_time = time

            self.run_time = time - self.start_time

            # Create a new DataFrame
            new_data = pd.DataFrame(
                {
                    MonitorColumns.TIME.value: [time],
                    MonitorColumns.CPU.value: [cpu],
                    MonitorColumns.MEMORY.value: [memory],
                    MonitorColumns.IO_RKB.value: [io_rkb],
                    MonitorColumns.IO_WKB.value: [io_wkb],
                    MonitorColumns.IOWAIT.value: [iowait],
                }
            )

            # Drop empty or all-NA columns
            new_data = new_data.dropna(axis=1, how="all")

            # Ensure the DataFrame is not empty after dropping columns
            if not new_data.empty:
                if self.df.empty:
                    self.df = new_data
                else:
                    self.df = pd.concat([self.df, new_data], ignore_index=True)


def performance(run_id, session=None, output_dir=None, output_prefix="") -> pd.DataFrame:
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)

    total_performance_df = pd.DataFrame()

    # Get run info from omics
    if session:
        omics_client = session.client("omics")
        logs_client = session.client("logs")
    else:
        omics_client = boto3.client("omics")
        logs_client = boto3.client("logs")
    run = get_run_info(run_id, client=omics_client)

    scattered_tasks = set()
    monitor_logs = []

    # Process monitor log for each task
    for task in run["tasks"]:
        print("------------------------------------------")
        print(f"Process monitor log for task {task['name']} (taskId: {task['taskId']})")
        monitor_log = process_monitor_log(run_id, task, client=logs_client)
        monitor_logs.append(monitor_log)

        new_row = pd.DataFrame(
            {
                "task": [task["name"]],
                "total_CPU": [monitor_log.total_cpu],
                "mean_%_CPU": [monitor_log.df[MonitorColumns.CPU.value].mean()],
                "max_%_CPU": [monitor_log.df[MonitorColumns.CPU.value].max()],
                "total_Memory(GB)": [monitor_log.total_memory],
                "mean_%_Memory": [monitor_log.df[MonitorColumns.MEMORY.value].mean()],
                "max_%_Memory": [monitor_log.df[MonitorColumns.MEMORY.value].max()],
                "run_time (hours)": [monitor_log.run_time.total_seconds() / 3600],
                "mean_IO_rKb/s": [monitor_log.df[MonitorColumns.IO_RKB.value].mean()],
                "max_IO_rKb/s": [monitor_log.df[MonitorColumns.IO_RKB.value].max()],
                "mean_IO_wKb/s": [monitor_log.df[MonitorColumns.IO_WKB.value].mean()],
                "max_IO_wKb/s": [monitor_log.df[MonitorColumns.IO_WKB.value].max()],
                "mean_IOWait": [monitor_log.df[MonitorColumns.IOWAIT.value].mean()],
                "max_IOWait": [monitor_log.df[MonitorColumns.IOWAIT.value].max()],
                "count_entries": [monitor_log.df[MonitorColumns.CPU.value].count()],
            }
        )
        total_performance_df = pd.concat([total_performance_df, new_row], ignore_index=True)

        # check if the task is a scattered task
        if "-" in task["name"]:
            scattered_tasks.add(task["name"].split("-")[0])

    # Save performance data
    output = f"{output_prefix}omics_{run_id}.performance.csv"
    if output_dir is not None:
        output = f"{output_dir}/{output}"
    print(f"Saving performance data to: {output}")
    total_performance_df.to_csv(output, index=False)

    # Save figures to HTML
    save_figures_to_html(monitor_logs, run_id, scattered_tasks, output_dir, output_prefix)

    return total_performance_df


def process_monitor_log(run_id, task, client=None) -> MonitorLog:
    if not client:
        client = boto3.client("logs")

    task_id = task["taskId"]

    log_stream_name = f"run/{run_id}/task/{task_id}"
    log_group_name = OMICS_LOG_GROUP

    monitor_log = MonitorLog()
    monitor_log.task_name = task["name"]
    monitor_log.task_id = task_id

    print(f"Get log events for log group '{log_group_name}' and log stream '{log_stream_name}'")

    # Get log events of specific task from CloudWatch that contain the word 'MONITORING'
    response = client.filter_log_events(
        logGroupName=log_group_name, logStreamNames=[log_stream_name], filterPattern="MONITORING"
    )

    # check that log is not empty
    if not response.get("events"):
        print(f"No events found for log group '{log_group_name}' and log stream '{log_stream_name}'")
        return monitor_log

    # process monitoring log events
    for event in response["events"]:
        monitor_log.process_line(event["message"])

    # get next page of log events
    while len(response.get("events")) > 0 and response.get("nextToken"):
        response = client.filter_log_events(
            logGroupName=log_group_name,
            logStreamNames=[log_stream_name],
            filterPattern="MONITORING",
            nextToken=response.get("nextToken"),
        )

        for event in response["events"]:
            monitor_log.process_line(event["message"])

    print(f"Done processing monitor log for task {task_id}")
    return monitor_log


def save_figures_to_html(monitor_logs, run_id, scattered_tasks, output_dir=None, output_prefix=""):  # noqa: C901, PLR0912, PLR0915 #TODO: refactor this function
    print("Generate and save performance plots to HTML...")
    dont_plot_tasks = []
    plot_tasks = []

    for monitor_log in monitor_logs:
        # Monitor log script prints every 10 seconds.
        # If monitor_log.df has less than 30 entires it means it worked less than 5 minutes and we don't need
        # the plot for this task.
        if monitor_log.df.shape[0] < 30:  # noqa: PLR2004
            print(
                f"Task: {monitor_log.task_name} (id:{monitor_log.task_id}) run less than 5 minutes "
                "and will not be plotted"
            )
            dont_plot_tasks.append(monitor_log)
        elif not any(
            task in monitor_log.task_name for task in scattered_tasks
        ):  # check if the task name is not a subtring of a scattered task
            plot_tasks.append(monitor_log)

    # TODO: fix this logic to use 5 columns only if scattered tasks have IO metrics
    if len(scattered_tasks) > 0:
        if add_io_plots(plot_tasks):
            cols = 5  # for scattered tasks plot cpu, memory, IO_read, IO_write, IO_wait
        else:
            cols = 2  # for scattered tasks plot cpu and memory
    else:
        cols = 1  # no scattered tasks, each row will contain one plot

    # Combine all figures into one HTML file
    rows = len(plot_tasks) + len(scattered_tasks)  # Each figure will be in its own row
    if rows == 0:
        print("No tasks to plot")
        return

    # regular tasks span over all columns, for the scattered tasks, each coulmn will have a plot
    specs = []
    for _ in range(len(plot_tasks)):
        # E.g., if col=5, it will result to [{'colspan': 5}, None, None, None, None] for each row,
        # meaning the first column will span all 5 columns
        specs.append([{"colspan": cols}] + [None] * (cols - 1))
    for _ in range(len(scattered_tasks)):
        # E.g., if col=5, it will result to [{}, {}, {}, {}, {}] for each row, meaning each column will have a plot
        specs.append([{} for _ in range(cols)])

    subplot_titles = [f"Task: {monitor_log.task_name} (id:{monitor_log.task_id})" for monitor_log in plot_tasks]
    combined_fig = make_subplots(rows=rows, cols=cols, specs=specs, subplot_titles=subplot_titles)

    add_to_legend = set()

    for i, monitor_log in enumerate(plot_tasks, start=1):
        print(f"Plotting task: {monitor_log.task_name} (id:{monitor_log.task_id})")
        # Add traces for each metric
        add_trace(combined_fig, monitor_log, MonitorColumns.CPU.value, i)
        add_trace(combined_fig, monitor_log, MonitorColumns.MEMORY.value, i)
        # Add IO metrics only if the coresepong coulmns are not empty and sum is not 0
        if (
            monitor_log.df[MonitorColumns.IO_RKB.value].notna().any()
            and monitor_log.df[MonitorColumns.IO_RKB.value].sum() != 0
        ):
            add_trace(combined_fig, monitor_log, MonitorColumns.IO_RKB.value, i, normelize=True)
            add_to_legend.add(MonitorColumns.IO_RKB.value)
        if (
            monitor_log.df[MonitorColumns.IO_WKB.value].notna().any()
            and monitor_log.df[MonitorColumns.IO_WKB.value].sum() != 0
        ):
            add_trace(combined_fig, monitor_log, MonitorColumns.IO_WKB.value, i, normelize=True)
            add_to_legend.add(MonitorColumns.IO_WKB.value)
        if (
            monitor_log.df[MonitorColumns.IOWAIT.value].notna().any()
            and monitor_log.df[MonitorColumns.IOWAIT.value].sum() != 0
        ):
            add_trace(combined_fig, monitor_log, MonitorColumns.IOWAIT.value, i)
            add_to_legend.add(MonitorColumns.IOWAIT.value)

        # Add caption for each subplot
        combined_fig.add_annotation(
            text=(
                f"#CPU: {monitor_log.total_cpu}, Memory(Gib): {monitor_log.total_memory}, "
                f"runtime(H): {monitor_log.run_time.total_seconds()/3600:.2f}"
            ),
            xref=f"x{i if i != 1 else ''} domain",  # Reference the x-axis of the ith subplot
            yref=f"y{i if i != 1 else ''} domain",  # Reference the y-axis of the ith subplot
            x=0.5,  # Position the caption in the middle of the subplot horizontally
            y=-0.15,  # Position the caption just below the subplot
            showarrow=False,
            font={"size": 12},
            align="center",
            xanchor="center",
            yanchor="top",
        )

    # Add missing lables in the legend by adding invisible traces
    for label in add_to_legend:
        combined_fig.add_trace(
            go.Scatter(x=[None], y=[None], mode="lines", name=label),
            row=1,
            col=1,  # Add it to the first subplot where the leggend appears
        )

    # Add plots for scattered tasks
    for i, task in enumerate(scattered_tasks, start=len(plot_tasks) + 1):
        print(f"Plotting scattered task: {task}")
        scattered_task_monitor_logs = [log for log in monitor_logs if task in log.task_name]

        # Itereate over tasks within a single scattered task and add metrics to the plots
        for _, monitor_log in enumerate(scattered_task_monitor_logs, start=1):
            add_trace(
                combined_fig,
                monitor_log,
                MonitorColumns.CPU.value,
                row=i,
                col=1,
                normelize_time=True,
                name=monitor_log.task_name,
            )
            add_trace(
                combined_fig,
                monitor_log,
                MonitorColumns.MEMORY.value,
                row=i,
                col=2,
                normelize_time=True,
                name=monitor_log.task_name,
            )

            if (
                monitor_log.df[MonitorColumns.IO_RKB.value].notna().any()
                and monitor_log.df[MonitorColumns.IO_RKB.value].sum() != 0
            ):
                add_trace(
                    combined_fig,
                    monitor_log,
                    MonitorColumns.IO_RKB.value,
                    row=i,
                    col=3,
                    normelize=True,
                    normelize_time=True,
                    name=monitor_log.task_name,
                )

            if (
                monitor_log.df[MonitorColumns.IO_WKB.value].notna().any()
                and monitor_log.df[MonitorColumns.IO_WKB.value].sum() != 0
            ):
                add_trace(
                    combined_fig,
                    monitor_log,
                    MonitorColumns.IO_WKB.value,
                    row=i,
                    col=4,
                    normelize=True,
                    normelize_time=True,
                    name=monitor_log.task_name,
                )

            if (
                monitor_log.df[MonitorColumns.IOWAIT.value].notna().any()
                and monitor_log.df[MonitorColumns.IOWAIT.value].sum() != 0
            ):
                add_trace(
                    combined_fig,
                    monitor_log,
                    MonitorColumns.IOWAIT.value,
                    row=i,
                    col=5,
                    normelize=False,
                    normelize_time=True,
                    name=monitor_log.task_name,
                )

        # Add caption for each subplot
        cpu = scattered_task_monitor_logs[0].total_cpu
        memory = scattered_task_monitor_logs[0].total_memory
        average_runtime = (
            sum([monitor_log.run_time.total_seconds() for monitor_log in scattered_task_monitor_logs])
            / len(scattered_task_monitor_logs)
            / 3600
        )
        combined_fig.add_annotation(
            text=(
                f"Scattered Task: {task} #CPU: {cpu}, Memory(Gib): {memory}, "
                f"average runtime(H): {average_runtime:.2f}"
            ),
            xref="paper",
            yref="paper",
            x=0.5,  # Middle of the row
            y=(1 - (i - 0.8) / rows),  # Top of the row
            showarrow=False,
            font={"size": 12},
            align="center",
            xanchor="center",
            yanchor="auto",
        )

    # touch up the layout and add title and subtext
    combined_fig.update_yaxes(title="%", range=[0, 110])
    combined_fig.update_layout(
        height=300 * rows + 50, title_text=f"Omics {run_id} Performance Plots", title_font={"size": 20}
    )
    if len(dont_plot_tasks) > 0:
        task_runtimes = [
            f"{monitor_log.task_name} (runtime (H): {monitor_log.run_time.total_seconds()/3600:.2f})"
            for monitor_log in dont_plot_tasks
        ]
        combined_fig.add_annotation(
            text=(f"Short tasks without a plot: {task_runtimes}"),
            xref="paper",  # Position relative to the entire plotting area
            yref="paper",  # Position relative to the entire plotting area
            x=0,  # Center the text horizontally
            y=1,  # Adjust this value as needed to position below the title
            showarrow=False,
            font={"size": 12},
            yanchor="bottom",
            yshift=25,  # Shift down by 25 pixels to add more "padding"
        )

    # Save the combined figure to an HTML file
    plots_dir = PLOTS_DIR.replace(RUN_ID_PLACEHOLDER, run_id)
    if output_dir is not None:
        plots_dir = f"{output_dir}/{plots_dir}"
    os.makedirs(plots_dir, exist_ok=True)
    combined_fig_output = f"{plots_dir}/{output_prefix}omics_{run_id}_performance_plots.html"
    print(f"Saving performance plots report to: {combined_fig_output}")
    combined_fig.write_html(combined_fig_output)


def add_io_plots(monitor_logs) -> bool:
    for monitor_log in monitor_logs:
        if (
            monitor_log.df[MonitorColumns.IO_RKB.value].notna().any()
            and monitor_log.df[MonitorColumns.IO_RKB.value].sum() != 0
        ):
            return True
        if (
            monitor_log.df[MonitorColumns.IO_WKB.value].notna().any()
            and monitor_log.df[MonitorColumns.IO_WKB.value].sum() != 0
        ):
            return True
        if (
            monitor_log.df[MonitorColumns.IOWAIT.value].notna().any()
            and monitor_log.df[MonitorColumns.IOWAIT.value].sum() != 0
        ):
            return True
    return False


def add_trace(combined_fig, monitor_log, col_name, row, col=1, name=None, *, normelize=False, normelize_time=False):
    show_legend = row == 1

    if normelize:
        monitor_log.df[col_name] = monitor_log.df[col_name] / monitor_log.df[col_name].max() * 100
    if normelize_time:
        monitor_log.df[MonitorColumns.TIME.value] = (
            monitor_log.df[MonitorColumns.TIME.value] - monitor_log.df[MonitorColumns.TIME.value].min()
        )

    if not name:
        name = col_name

    combined_fig.add_trace(
        go.Scatter(
            x=monitor_log.df[MonitorColumns.TIME.value],
            y=monitor_log.df[col_name],
            mode="lines",
            name=name,
            line={"color": monitor_log.colors[col_name]},
            showlegend=show_legend,
        ),
        row=row,
        col=col,
    )


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--region", type=str, help="AWS region to use", default="us-east-1")
    parser.add_argument("--run-id", type=str, help="HealthOmics workflow run-id to analyze")
    parser.add_argument("--output-path", type=str, help="Output dir to save performance data", required=False)
    parser.add_argument("--output-prefix", type=str, help="File name prefix for the output", required=False, default="")

    args = parser.parse_args()
    session = boto3.Session(region_name=args.region)

    performance(args.run_id, session=session, output_dir=args.output_path, output_prefix=args.output_prefix)



"""
Command line access to PAPYRUS mongoDB.
See documentation in
https://ultimagen.atlassian.net/wiki/spaces/AG/pages/1428914739/Papyrus+Metrics+Infrastructure+Proof+of+Concept#Command-Line-Access
"""

import json
import os
import warnings
from enum import Enum
from typing import Any

import pandas as pd
import pymongo


class Collections(Enum):
    CROMWELL = "pipelines"
    RUNS = "runs"
    SAMPLES = "samples"
    EXECUTIONS = "executions"
    PPMSEQ = "ppmseq"


def set_papyrus_access():
    """Disables/Enables papyrus access depends on PAPYRUS_ACCESS_STRING env"""
    disable_papyrus_access = False
    collections = {}
    if "PAPYRUS_ACCESS_STRING" in os.environ:
        my_client = initialize_client()
        my_db = my_client["pipelines"]
        collections[Collections.CROMWELL] = my_db["pipelines"]
        collections[Collections.RUNS] = my_db["runs"]
        collections[Collections.EXECUTIONS] = my_db["runs.executions"]
        collections[Collections.SAMPLES] = my_db["runs.executions.samples"]
        collections[Collections.PPMSEQ] = my_db["ppmseq_workflows_view"]
    else:
        warnings.warn("Define PAPYRUS_ACCESS_STRING environmental variable to enable access to Papyrus", stacklevel=2)
        warnings.warn(
            "Example: export PAPYRUS_ACCESS_STRING=mongodb+srv://[user]:[passwd]@testcluster.jm2x3.mongodb.net/test",
            stacklevel=2,
        )
        disable_papyrus_access = True
    return disable_papyrus_access, collections


def initialize_client() -> pymongo.MongoClient:
    """Initializes pymongo client with the access string that is read from PAPYRUS_ACCESS_STRING
    environmental variable.

    Parameters:
    -----------
    None

    Returns
    -------
    pymongo.MongoClient
    """

    myclient = pymongo.MongoClient(os.environ["PAPYRUS_ACCESS_STRING"])
    return myclient


def query_database(query: dict, collection: str = "pipelines", **kwargs: Any) -> list:
    """Querying pipelines database. For easy access
    Define PAPYRUS_ACCESS_STRING environmental variable
    Example: export PAPYRUS_ACCESS_STRING=mongodb+srv://[user]:[passwd]@testcluster.jm2x3.mongodb.net/test

    Parameters
    ----------
    query : dict
        Pymongo query (dictionary)
    collection: str
        Supported - 'pipelines' (default), 'runs', 'executions', samples', 'ppmseq'
    **kwargs: Any
        kwargs to pass to pymongo.find

    Returns
    -------
    list
        List of documents
    """
    disable_papyrus_access, collections = set_papyrus_access()
    assert not disable_papyrus_access, "Database access not available through PAPYRUS_ACCESS_STRING"  # noqa s101
    return list(collections[Collections(collection)].find(query, **kwargs))


DEFAULT_METRICS_TO_REPORT = [
    "AlignmentSummaryMetrics",
    "Contamination",
    "DuplicationMetrics",
    "GcBiasDetailMetrics",
    "GcBiasSummaryMetrics",
    "QualityYieldMetrics",
    "RawWgsMetrics",
    "WgsMetrics",
    "stats_coverage",
    "short_report_/all_data",
    "short_report_/all_data_gt",
]


def metrics2df(doc: dict, metrics_to_report: list | None = None) -> pd.DataFrame:
    """Converts metrics document to pandas dataframe

    Parameters
    ----------
    doc: dict
        Single document from mongodb (dictionary, output of query_database)
    metrics_to_report: list, optional
        which metrics should be reported (default {", ".join(DEFAULT_METRICS_TO_REPORT)})

    Returns
    -------
    pd.DataFrame
        1xn dataframe with two level index on columns: (metric_type, metric_name)
    """

    if metrics_to_report is None:
        metrics_to_report = DEFAULT_METRICS_TO_REPORT

    if "workflowEntity" not in doc["metadata"]:  # omics documents
        metadata = pd.DataFrame((pd.DataFrame(_cleanup_metadata(doc["metadata"]))).loc["submission"]).T

    else:
        metadata = pd.DataFrame(
            (pd.DataFrame(_cleanup_metadata(doc["metadata"])))
            .query('(workflowEntity=="sample") | (workflowEntity=="Sample") | (workflowEntity=="Unknown")')
            .loc["entityType"]
        ).T
    metadata.index = pd.Index([0])
    metadata = pd.concat({"metadata": metadata}, axis=1)
    result = [
        (x, pd.read_json(json.dumps(doc["metrics"][x]), orient="table"))
        for x in doc["metrics"]
        if x in metrics_to_report
    ]
    result = [x for x in result if x[1].shape[0] == 1]
    result_df = pd.concat((metadata, pd.concat(dict(result), axis=1)), axis=1).set_index(("metadata", "workflowId"))
    result_df.index = result_df.index.rename("workflowId")
    return result_df


def inputs2df(doc: dict) -> pd.DataFrame:
    """Returns a dataframe of inputs and outputs metadata

    Parameters
    ----------
    doc: dict
        Single document from mongoDB

    Returns
    -------
    pd.DataFrame
        Dataframe of inputs and outputs combined (single row)
    """
    if "workflowEntity" not in doc["metadata"]:  # omics documents
        metadata = pd.DataFrame(_cleanup_metadata(doc["metadata"])).loc["submission"]
    else:
        metadata = (
            pd.DataFrame(_cleanup_metadata(doc["metadata"]))
            .query('(workflowEntity=="sample") | ' + '(workflowEntity=="Sample") | ' + '(workflowEntity=="Unknown")')
            .loc["entityType"]
        )
    inputs = pd.Series(doc["inputs"])
    outputs = pd.Series(doc["outputs"])
    return pd.DataFrame(pd.concat((metadata, inputs, outputs))).T.set_index("workflowId")


def nexus_metrics_to_df(input_dict: dict) -> pd.DataFrame:
    """Returns a dataframe of nexus metrics from run/execution/sample documentation

    Parameters
    ----------
    input_dict: dict
        Dictionary from mongoDB

    Returns
    -------
    pd.DataFrame
        Dataframe of parameters
    """
    s = pd.Series(input_dict).drop("_id")
    values = s.to_numpy()
    index = pd.MultiIndex.from_tuples([x.split("_", maxsplit=1) for x in s.index])
    name = s["metadata_sequencingRunId"]
    return pd.DataFrame(columns=[name], index=index, data=values).T


def _cleanup_metadata(input_dict: dict) -> dict:
    """Cleans up metadata - removes empty lists

    Parameters
    ----------
    input_dict: dict
        Input metadata dict

    Returns
    -------
    dict
        Same dictionary without values of type lists (confuses dataframe conversion)
    """
    for k in list(input_dict.keys()):
        if isinstance(input_dict[k], list):
            del input_dict[k]
    return input_dict



import os
from enum import Enum

import boto3
import botocore
import pandas as pd

from ugbio_omics.get_run_info import get_run_info

RUN_ID_PLACEHOLDER = "<RUN_ID>"
PLOTS_DIR = f"omics_{RUN_ID_PLACEHOLDER}_plots"


class Columns(Enum):
    ESTIMATED_USD_COLUMN = "estimatedUSD"
    TYPE_COLUMN = "type"
    NAME_COLUMN = "name"
    OMICS_INSTANCE_TYPE_RESERVED = "omicsInstanceTypeReserved"
    CPU_REQUESTED = "cpusRequested"
    MEMORY_REQUESTED_GIB = "memoryRequestedGiB"
    GPUS_REQUESTED = "gpusRequested"
    RUNNING_SECONDS = "runningSeconds"


class RunCost:
    def __init__(self, run_id, output_dir=None, output_prefix="", session=None):
        self.run_id = run_id
        self.output_dir = output_dir
        if self.output_dir:
            os.makedirs(output_dir, exist_ok=True)
        self.output_prefix = output_prefix
        self.session = session
        self.cost_csv = None
        self.cost_df = None

    def calculate_run_cost(self) -> None:
        cost_output = f"{self.output_prefix}omics_{self.run_id}.cost.csv"
        plots_dir = PLOTS_DIR.replace(RUN_ID_PLACEHOLDER, self.run_id)

        if self.output_dir:
            cost_output = f"{self.output_dir}/{cost_output}"
            plots_dir = f"{self.output_dir}/{plots_dir}"

        os.makedirs(plots_dir, exist_ok=True)

        if self.session:
            omics_client = self.session.client("omics")
            s3_client = self.session.client("s3")
        else:
            omics_client = boto3.client("omics")
            s3_client = boto3.client("s3")

        # Get bucket name and path for run outputs
        run = get_run_info(self.run_id, client=omics_client, get_tasks=False)
        bucket_name, path = self.get_logs_key(run)

        # Download cost.csv from s3 run outputs
        key = f"{path}/run-{self.run_id}.csv"
        cost_csv_uri = f"s3://{bucket_name}/{key}"
        print(f"Run cost downaloaded from {cost_csv_uri} and saved to {cost_output}")
        with open(cost_output, "wb") as f:
            s3_client.download_fileobj(bucket_name, key, f)

        # Download timeline plot from s3 run outputs (if avaliable)
        key = f"{path}/plots/{self.run_id}_timeline.html"
        timeline_plots_uri = f"s3://{bucket_name}/{key}"
        plots_output = f"{plots_dir}/{self.run_id}_timeline.html"
        print(f"Timeline plot downaloaded from {timeline_plots_uri} and saved to {plots_output}")
        with open(plots_output, "wb") as f:
            try:
                s3_client.download_fileobj(bucket_name, key, f)
            except botocore.exceptions.ClientError as e:
                print(f"Error downloading timeline plot: {e}")
                print(f"Skipping timeline plot download for run {self.run_id}")

        self.cost_csv = cost_output
        self.cost_df = pd.read_csv(self.cost_csv)
        # make sure cost and duration are numeric
        for col in [Columns.ESTIMATED_USD_COLUMN.value, Columns.RUNNING_SECONDS.value]:
            self.cost_df[col] = pd.to_numeric(self.cost_df[col], errors="coerce")

    def get_logs_key(self, run):
        run_output_uri = run["outputUri"]
        bucket_name, pipeline = run_output_uri.replace("s3://", "").split("/", 1)
        path = f"{pipeline}/{self.run_id}/logs"
        return bucket_name, path

    def get_total_cost(self) -> float:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[Columns.ESTIMATED_USD_COLUMN.value].sum()

    def get_total_runtime(self) -> float:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[self.cost_df[Columns.TYPE_COLUMN.value] == "run"][Columns.RUNNING_SECONDS.value].to_numpy()[
            0
        ]

    def get_storage_cost(self) -> float:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[self.cost_df[Columns.TYPE_COLUMN.value] == "run"][
            Columns.ESTIMATED_USD_COLUMN.value
        ].to_numpy()[0]

    def get_tasks_cost(self) -> pd.DataFrame:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[self.cost_df[Columns.TYPE_COLUMN.value] == "task"][
            [
                Columns.NAME_COLUMN.value,
                Columns.ESTIMATED_USD_COLUMN.value,
                Columns.OMICS_INSTANCE_TYPE_RESERVED.value,
                Columns.RUNNING_SECONDS.value,
            ]
        ]

    def get_tasks_resources(self) -> pd.DataFrame:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[self.cost_df[Columns.TYPE_COLUMN.value] == "task"][
            [
                Columns.NAME_COLUMN.value,
                Columns.OMICS_INSTANCE_TYPE_RESERVED.value,
                Columns.CPU_REQUESTED.value,
                Columns.MEMORY_REQUESTED_GIB.value,
                Columns.GPUS_REQUESTED.value,
            ]
        ]

    def get_run_cost(self) -> pd.DataFrame:
        tasks_cost_df = self.get_tasks_cost()

        # group by task name, calculate sum of cost and mean of runtime
        tasks_cost_df[Columns.NAME_COLUMN.value] = tasks_cost_df[Columns.NAME_COLUMN.value].str.split("-").str[0]
        grouped_df = (
            tasks_cost_df.groupby(Columns.NAME_COLUMN.value)
            .agg(
                {
                    Columns.ESTIMATED_USD_COLUMN.value: "sum",
                    Columns.RUNNING_SECONDS.value: "mean",
                    Columns.OMICS_INSTANCE_TYPE_RESERVED.value: "first",
                }
            )
            .reset_index()
        )

        # add storage and total cost
        storage_total_df = pd.DataFrame(
            {
                Columns.NAME_COLUMN.value: ["storage", "total"],
                Columns.ESTIMATED_USD_COLUMN.value: [self.get_storage_cost(), self.get_total_cost()],
                Columns.RUNNING_SECONDS.value: [None, self.get_total_runtime()],
            }
        )

        grouped_df = pd.concat([grouped_df, storage_total_df], ignore_index=True)

        # convert runtime in seconds to hours
        grouped_df[Columns.RUNNING_SECONDS.value] = (
            grouped_df[Columns.RUNNING_SECONDS.value] / 60 / 60
        )  # change runtime to hours

        # Rename columns
        grouped_df = grouped_df.rename(
            columns={
                Columns.NAME_COLUMN.value: "task",
                Columns.ESTIMATED_USD_COLUMN.value: "cost",
                Columns.OMICS_INSTANCE_TYPE_RESERVED.value: "instance",
                Columns.RUNNING_SECONDS.value: "run_time (hours)",
            }
        )
        return grouped_df
